import contextvars
import functools
from asyncio import CancelledError, events
from concurrent.futures import ThreadPoolExecutor
from typing import Callable


async def to_thread_in_pool(pool: ThreadPoolExecutor, func: Callable, /, *args, **kwargs):
    """Copy of asyncio.to_thread, but:
    - using a custom thread pool
    - not requiring Python 3.9
    - cancelling the future when the task is cancelled

    Asynchronously run function *func* in a separate thread.

    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    Return a coroutine that can be awaited to get the eventual result of *func*.
    """
    loop = events.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    future = loop.run_in_executor(pool, func_call)
    try:
        return await future
    except CancelledError:
        future.cancel()  # TODO: this is just a precaution, should not be needed, remove?
        raise



from rate_limited.resources import Resource


def get_requests_per_minute(quota: int) -> Resource:
    return Resource(
        name="requests_per_minute",
        quota=quota,
        time_window_seconds=60,
        arguments_usage_extractor=lambda _: 1,
    )



from collections import deque
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import Any, Callable, Optional

from rate_limited.calls import Call

Unit = float


@dataclass
class UsageLog:
    timestamp: datetime
    amount: Unit


class Resource:
    def __init__(
        self,
        name: str,
        quota: Unit,
        time_window_seconds: float,
        arguments_usage_extractor: Optional[Callable[[Call], Unit]] = None,
        results_usage_extractor: Optional[Callable[[Any], Unit]] = None,
        max_results_usage_estimator: Optional[Callable[[Call], Unit]] = None,
    ):
        """
        Defines a resource

        Args:
            name: name of the resource
            quota: maximum amount of the resource that can be used in the time window
            time_window_seconds: time window in seconds
            arguments_usage_extractor: function that extracts the amount of resource used from
                the arguments
            results_usage_extractor: function that extracts the amount of resource used from
                the results
            max_results_usage_estimator: function that extracts an upper bound on the amount of
                resource that might be used when results are returned, based on the arguments
                (this is used to pre-allocate usage, pre-allocation is then replaced with the
                actual usage when the results are returned). Only used in combination with
                results_usage_extractor.
        """
        self.name = name
        self.quota = quota
        self.time_window_seconds = time_window_seconds
        self._used = Unit(0)
        self.usage_log: deque = deque()
        self._reserved = Unit(0)

        self.arguments_usage_extractor = arguments_usage_extractor
        self.results_usage_extractor = results_usage_extractor
        self.max_results_usage_estimator = max_results_usage_estimator  # TODO: consider renaming

        if self.max_results_usage_estimator and not self.results_usage_extractor:
            raise ValueError(
                "max_results_usage_estimator can only be used when results_usage_extractor is "
                "also provided"
            )

    def __repr__(self):
        return (
            f"Resource: {self.name} - available: {self.get_amount_available()}/{self.quota}"
            f" (used: {self._used}, reserved: {self._reserved})"
        )

    def add_usage(self, amount: Unit) -> None:
        self._used += amount
        self.usage_log.append(UsageLog(datetime.now(), amount))

    def reserve_amount(self, amount: Unit) -> None:
        self._reserved += amount

    def remove_reserved(self, amount: Unit) -> None:
        self._reserved -= amount

    def get_amount_used(self) -> Unit:
        """
        Returns the amount used in the last time_window_seconds - ignoring reserved amounts.
        """
        while self.usage_log and (
            (datetime.now() - self.usage_log[0].timestamp).seconds > self.time_window_seconds
        ):
            self._used -= self.usage_log.popleft().amount
        return self._used

    def get_amount_available(self) -> Unit:
        """
        Returns the amount available - taking into account reserved amounts.
        """
        return self.quota - self.get_amount_used() - self._reserved

    def is_available(self, amount) -> bool:
        """
        Returns True if there is enough remaining quota to use the given amount.
        """
        return self.get_amount_available() >= amount

    def get_next_expiration(self) -> datetime:
        """
        Returns the timestamp of the next expiration of a usage log. Returns current time
        if there are no usage logs.
        """
        if not self.usage_log:
            return datetime.now()
        return self.usage_log[0].timestamp + timedelta(seconds=self.time_window_seconds)



from asyncio import Queue
from typing import Any


class CompletionTrackingQueue(Queue):
    def __init__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize=maxsize)
        self.all_tasks_count = 0
        self.tasks_done_count = 0

    def put_nowait(self, item: Any) -> None:
        super().put_nowait(item)
        self.all_tasks_count += 1

    def task_done(self) -> None:
        super().task_done()
        self.tasks_done_count += 1

    def all_tasks_done(self) -> bool:
        """
        Return True if the queue is empty and task_done() was called for every item that had been
        put() into the queue.

        This is different than Queue.empty() which returns True if each put() was matched by
        a get(), but might still be waiting for task_done() to be called.
        """
        return self.all_tasks_count == self.tasks_done_count



from dataclasses import dataclass, field
from functools import cached_property
from inspect import signature
from typing import Any, Callable, List


@dataclass
class Call:
    function: Callable
    args: tuple
    kwargs: dict
    num_retries: int = 0
    result: Any = None
    exceptions: List[Exception] = field(default_factory=list)

    @cached_property
    def all_arguments_dict(self) -> dict:
        all_args_dict = signature(self.function).bind(*self.args, **self.kwargs).arguments
        kwargs = all_args_dict.pop("kwargs", {})
        return {**all_args_dict, **kwargs}

    @cached_property
    def default_params(self) -> dict:
        return {
            name: param.default
            for name, param in signature(self.function).parameters.items()
            if param.default is not param.empty
        }

    def get_argument(self, name: str) -> Any:
        """
        Returns the value of the argument with the given name - whether it was passed positionally,
        as a keyword argument, or as a default value.
        """
        return self.all_arguments_dict.get(name) or self.default_params.get(name)



from typing import Optional


class ProgressBar:
    """A progress bar that uses tqdm if available - does nothing otherwise."""

    def __init__(self, total: int):
        try:
            from tqdm import tqdm

            self.pbar: Optional[tqdm] = tqdm(total=total)
        except ImportError:
            self.pbar = None

    def set_state(self, num_completed: int, num_total: Optional[int] = 0):
        if self.pbar:
            self.pbar.update(num_completed - self.pbar.n)
            if num_total is not None:
                old_total = self.pbar.total
                self.pbar.total = num_total
                if old_total != num_total:
                    self.pbar.refresh()
        # TODO: optional text-based logging, perhaps at DEBUG level

    def close(self):
        if self.pbar:
            self.pbar.close()



import asyncio
import traceback
from asyncio import create_task, gather
from asyncio import sleep as asyncio_sleep
from concurrent.futures import ThreadPoolExecutor
from logging import getLogger
from typing import Callable, Collection, List, Optional, Tuple

from rate_limited.calls import Call
from rate_limited.progress_bar import ProgressBar
from rate_limited.queue import CompletionTrackingQueue
from rate_limited.resource_manager import ResourceManager
from rate_limited.resources import Resource
from rate_limited.threading_utils import to_thread_in_pool


class Runner:
    def __init__(
        self,
        function: Callable,
        resources: Collection[Resource],
        max_concurrent: int,
        max_retries: int = 5,
        progress_interval: float = 1.0,
    ):
        self.function = function
        self.resource_manager = ResourceManager(resources)
        self.max_concurrent = max_concurrent
        self.requests_executor_pool = ThreadPoolExecutor(max_workers=max_concurrent)
        self.max_retries = max_retries
        self.progress_interval = progress_interval
        # TODO: add verification functions?
        # (checking if response meets criteria, retrying otherwise)

        self.logger = getLogger(f"rate_limited.Runner.{function.__name__}")

        # list of all calls that have been scheduled, in order of scheduling
        self.scheduled_calls: List[Call] = []
        # queue of calls to be executed in the current run()/run_coro() call, including retries
        # (needs to be initialized in the context of the event loop we will execute in)
        self.execution_queue: Optional[CompletionTrackingQueue] = None

        self.interrupted = False

    def schedule(self, *args, **kwargs) -> None:
        """
        Use to schedule a call to the function

        If the normal call looked liked `my_function("some text", temperature=0.5)`, then
        `runner.schedule("some text", temperature=0.5)` should be used instead.
        """
        # TODO: use docstring from self.function at runtime
        call = Call(self.function, args, kwargs)
        if not self.resource_manager.is_call_allowed(call):
            raise ValueError(f"Call {call} exceeds resource quota - will never be executed")
        self.scheduled_calls.append(call)
        if self.execution_queue is not None:
            self.execution_queue.put_nowait(call)

    def run(self) -> Tuple[list, list]:
        """
        Runs the scheduled calls, returning a tuple of:
        - results (list, in order of scheduling) and
        - exceptions(list of lists, in order of scheduling)

        Can be called from both sync and async code
        (so that the same code can be used in a script and a notebook - Jupyter runs an event loop)

        Handles KeyboardInterrupt (SIGINT) gracefully by returning partial results.
        """
        with ThreadPoolExecutor(1) as pool:
            future = pool.submit(self._run_sync)
            try:
                return future.result()
            except KeyboardInterrupt:
                self.logger.warning("Interrupted, collecting already returned results")
                self.interrupted = True
                return future.result()

    def _run_sync(self) -> Tuple[list, list]:
        """
        Execute run_coro() from sync code - starting a new event loop
        """
        return asyncio.run(self.run_coro())

    async def run_coro(self) -> Tuple[list, list]:
        """
        Coroutine actually implementing the execution of the scheduled calls

        Can be used directly from async code (e.g. in a Jupyter notebook):
        ```
        results, exceptions = await runner.run_coro()
        ```
        or indirectly via run() - from any environment:
        ```
        results, exceptions = runner.run()
        ```
        """
        self.interrupted = False
        self.initialize_in_event_loop()
        assert self.resource_manager.condition is not None  # for mypy's benefit
        assert self.execution_queue is not None

        worker_tasks = [create_task(self.worker()) for _ in range(self.max_concurrent)]

        pbar = ProgressBar(total=len(self.scheduled_calls))
        while not self.execution_queue.all_tasks_done() and not self.interrupted:
            num_completed = self.execution_queue.tasks_done_count
            self.logger.debug(f"Tasks done: {num_completed} interrupted: {self.interrupted}")
            pbar.set_state(num_completed, num_total=self.execution_queue.all_tasks_count)
            # TODO: use a condition/signal instead? though OS support varies
            # NB: KeyboardInterrupt handling will wait for this sleep too - should not be too long
            await asyncio_sleep(self.progress_interval)
            if not self.interrupted:
                # wake up waiting workers - perhaps now resources are available
                await self.resource_manager.notify_waiting()
        pbar.set_state(self.execution_queue.tasks_done_count)
        pbar.close()
        self.logger.debug("Done processing - cancelling worker tasks")
        for task in worker_tasks:
            task.cancel()
        await gather(*worker_tasks, return_exceptions=True)
        self.logger.info("Workers finished")
        results = [call.result for call in self.scheduled_calls]
        exception_lists = [call.exceptions for call in self.scheduled_calls]
        if not self.interrupted:
            self.scheduled_calls = []  # only clear the queue if returning complete results
        # TODO: consider returning a generator, instead of waiting for all calls to finish?
        return results, exception_lists

    def initialize_in_event_loop(self):
        """
        Needs to be called once we are in the context of the (current) event loop.

        Note: If called multiple times (via run()) the event loop will be different -
        we need to re-initialize the synchronization primitives.
        """
        # NB: (re-)initializing asyncio synchronization primitives not needed for newer Python
        # versions (>=3.10?) - but queue cleanup still needed
        self.resource_manager.initialize_in_event_loop()
        self.execution_queue = CompletionTrackingQueue()
        for call in self.scheduled_calls:
            if call.result is None and call.num_retries <= self.max_retries:
                self.execution_queue.put_nowait(call)

    async def worker(self):
        assert self.execution_queue is not None
        while not self.interrupted:
            # wait to get a task
            call = await self.execution_queue.get()
            # wait for resources to be available
            await self.resource_manager.wait_for_resources(call)

            # starting to execute - but first, register the usage
            self.resource_manager.register_call(call)
            self.resource_manager.pre_allocate(call)
            try:
                # TODO: add a timeout mechanism?
                call.result = await to_thread_in_pool(
                    self.requests_executor_pool, self.function, *call.args, **call.kwargs
                )
                # TODO: are there cases where we need to register result-based usage on error?
                # (one case: if we have user-defined verification functions)
                self.resource_manager.register_result(call.result)
            except Exception as e:
                will_retry = call.num_retries < self.max_retries
                self.logger.warning(
                    f"Exception occurred, will retry: {will_retry}\n{traceback.format_exc()}"
                )
                call.exceptions.append(e)
                if will_retry:
                    call.num_retries += 1
                    self.execution_queue.put_nowait(call)
            finally:
                self.resource_manager.remove_pre_allocation(call)
                self.execution_queue.task_done()



from functools import partial
from typing import List

from rate_limited.apis.common import get_requests_per_minute
from rate_limited.calls import Call
from rate_limited.resources import Resource


def openai_chat_resources(requests_per_minute, tokens_per_minute, model_max_len) -> List[Resource]:
    # getting too close to the limit results in some refusals (there are some differences between
    # in how the usage is tracked - hard to reverse-engineer exactly)
    # TODO: clean this up somehow? (getting a closer match to the behavior of the API)
    token_limit = int(tokens_per_minute * 0.95)
    return [
        get_requests_per_minute(requests_per_minute),
        get_tokens_per_minute(token_limit, model_max_len),
    ]


def get_tokens_per_minute(quota: int, model_max_len: int) -> Resource:
    estimator = partial(estimate_tokens_used, model_max_len=model_max_len)
    return Resource(
        name="tokens_per_minute",
        quota=quota,
        time_window_seconds=60,
        max_results_usage_estimator=estimator,
        results_usage_extractor=get_used_tokens,
    )


def get_used_tokens(results: dict) -> int:
    total_tokens = results.get("usage", {}).get("total_tokens", None)
    if total_tokens is None:
        raise ValueError("Could not find total_tokens in results")
    return total_tokens


def estimate_tokens_used(call: Call, model_max_len: int) -> int:
    """
    Estimate (upper bound) of how many tokens might be used by a request - before running it
    """
    max_tokens_arg = call.get_argument("max_tokens")
    n_arg = call.get_argument("n") or 1  # how many completions to return
    if max_tokens_arg is not None:
        max_tokens = estimate_input_tokens(call) + max_tokens_arg
    else:
        max_tokens = model_max_len
    return max_tokens * n_arg


def estimate_input_tokens(call: Call) -> int:
    parts = []
    for key in ["messages", "functions"]:
        array = call.get_argument(key) or []
        # fmt: off
        parts += [
            part
            for dct in array
            for part in dct.values()
            if isinstance(part, str)
        ]
        # fmt: on
    # very rough over-estimate - assuming 2 tokens per word
    tokens_uppoer_bound = sum(len(part.split(" ") * 2) for part in parts)
    # TODO: improve estimation - look into details of possible arguments
    # TODO: if installed, use tiktoken to count tokens exactly
    return tokens_uppoer_bound



import asyncio
from asyncio import Condition
from datetime import datetime
from logging import getLogger
from typing import Collection, Optional

from rate_limited.calls import Call
from rate_limited.resources import Resource, Unit


class ResourceManager:
    def __init__(self, resources: Collection[Resource]):
        self.resources = list(resources)
        self.condition: Optional[Condition] = None
        self.logger = getLogger("rate_limited.ResourceManager")

    def initialize_in_event_loop(self):
        if self.condition is None:
            self.condition = Condition()
            return

        current_loop = asyncio.get_running_loop()
        if self.condition._loop is not current_loop:  # type: ignore
            # new event loop -
            self.logger.debug("Detected event loop change - re-initializing condition")
            self.condition = Condition()

    def is_call_allowed(self, call: Call) -> bool:
        """
        Checks if the resources needed are below the quota - otherwise it will never be allowed
        """
        for resource in self.resources:
            needed = Unit(0)
            if resource.arguments_usage_extractor is not None:
                needed += resource.arguments_usage_extractor(call)
            if resource.max_results_usage_estimator is not None:
                needed += resource.max_results_usage_estimator(call)
            if needed > resource.quota:
                return False
        return True

    def register_call(self, call: Call):
        for resource in self.resources:
            if resource.arguments_usage_extractor:
                resource.add_usage(resource.arguments_usage_extractor(call))

    def pre_allocate(self, call: Call):
        for resource in self.resources:
            if resource.max_results_usage_estimator:
                resource.reserve_amount(resource.max_results_usage_estimator(call))

    def register_result(self, result):
        for resource in self.resources:
            if resource.results_usage_extractor:
                resource.add_usage(resource.results_usage_extractor(result))

    def remove_pre_allocation(self, call: Call):
        # Right now assuming that pre-allocation is only based on the call, this could change
        # to e.g. be also based on history of results
        for resource in self.resources:
            if resource.max_results_usage_estimator:
                resource.remove_reserved(resource.max_results_usage_estimator(call))

    def get_next_usage_expiration(self) -> datetime:
        return min(resource.get_next_expiration() for resource in self.resources)

    def _has_space_for_call(self, call: Call) -> bool:
        # important - we should NOT have any async code here!
        # (because we are inside a condition check)
        for resource in self.resources:
            needed = Unit(0)
            if resource.arguments_usage_extractor is not None:
                needed += resource.arguments_usage_extractor(call)
            if resource.max_results_usage_estimator is not None:
                needed += resource.max_results_usage_estimator(call)
            if not resource.is_available(needed):
                self.logger.debug(f"resource {resource.name} is not available: {resource}")
                return False
        self.logger.debug("all resources are available")
        return True

    async def wait_for_resources(self, call: Call):
        assert self.condition is not None
        async with self.condition:
            await self.condition.wait_for(lambda: self._has_space_for_call(call))

    async def notify_waiting(self):
        assert self.condition is not None
        async with self.condition:
            # TODO: this is too eager, we could only wake a subset of workers
            # (exact solution non-trivial?, gains likely negligible)
            self.condition.notify_all()
